/**
 * TypeScriptServerHost.ts
 */

import * as childProcess from "child_process"
import * as events from "events"
import * as fs from "fs"
import * as os from "os"
import * as path from "path"
import * as readline from "readline"

const tssPath = path.join(__dirname, "..", "..", "..", "..", "..", "node_modules", "typescript", "lib", "tsserver.js")

/**
 * Taken from definitions here:
 * https://github.com/Microsoft/TypeScript/blob/master/lib/protocol.d.ts#L5
 */
export interface ITextSpan {
    start: ILocation
    end: ILocation
}

export interface ILocation {
    line: number
    offset: number
}

export interface INavigationTree {
    text: string
    kind: string
    kindModifiers: string
    spans: ITextSpan[]
    childItems?: INavigationTree[]
}

/**
 * End definitions
 */

export class TypeScriptServerHost extends events.EventEmitter {

    private _tssProcess = null
    private _seqNumber = 0
    private _seqToPromises = {}
    private _rl: any

    public get pid(): number {
        return this._tssProcess.pid
    }

    constructor() {
        super()

        // Other tries for creating process:
        // this._tssProcess = childProcess.spawn("node", [tssPath], { stdio: "pipe", detached: true, shell: false });
        // this._tssProcess = childProcess.fork(tssPath, [], { stdio: "pipe "})
        //
        // On Windows, an 'npm' window would show up, so it seems like in this context,
        // exec was the most reliable

        // Note max buffer value - once this exceeded, the process will crash
        // TODO: Reload process, or looking at using the --eventPort option instead
        // This has some info on using eventPort: https://github.com/Microsoft/TypeScript/blob/master/src/server/server.ts
        // which might be more reliable
        // Can create the port using this here: https://github.com/Microsoft/TypeScript/blob/master/src/server/server.ts
        this._tssProcess = childProcess.exec("node " + tssPath, { maxBuffer: 500 * 1024 * 1024 }, (err) => {
            if (err) {
                console.error(err)
            }
        })
        console.log("Process ID: " + this._tssProcess.pid) // tslint:disable-line no-console

        this._rl = readline.createInterface({
            input: this._tssProcess.stdout,
            output: this._tssProcess.stdin,
            terminal: false,
        })

        this._tssProcess.stderr.on("data", (data, err) => {
            console.error("Error from tss: " + data)
        })

        this._tssProcess.on("error", (data) => {
            debugger // tslint:disable-line no-debugger
        })

        this._tssProcess.on("exit", (data) => {
            debugger // tslint:disable-line no-debugger
        })

        this._tssProcess.on("close", (data) => {
            debugger // tslint:disable-line no-debugger
        })

        this._rl.on("line", (msg) => {
            if (msg.indexOf("{") === 0) {
                this._parseResponse(msg)
            }
        })
    }

    public openFile(file: string): Promise<any> {
        return this._makeTssRequest("open", {
            file,
        })
    }

    public getProjectInfo(file: string): Promise<any> {
        return this._makeTssRequest("projectInfo", {
            file,
            needFileNameList: true,
        })
    }

    public getTypeDefinition(file: string, line: number, offset: number): Promise<void> {
        return this._makeTssRequest<void>("typeDefinition", {
            file,
            line,
            offset,
        })
    }

    public getFormattingEdits(file: string, line: number, offset: number, endLine: number, endOffset: number): Promise<any> {
        return this._makeTssRequest<void>("format", {
            file,
            line,
            offset,
            endLine,
            endOffset,
        })
    }

    public getCompletions(file: string, line: number, offset: number, prefix: string): Promise<any> {
        return this._makeTssRequest<void>("completions", {
            file,
            line,
            offset,
            prefix,
        })
    }

    public getCompletionDetails(file: string, line: number, offset: number, entryNames: string[]): Promise<any> {
        return this._makeTssRequest<void>("completionEntryDetails", {
            file,
            line,
            offset,
            entryNames,
        })
    }

    public updateFile(file: string, fileContent: string): Promise<void> {
        return this._makeTssRequest<void>("open", {
            file,
            fileContent,
        })
    }

    public getQuickInfo(file: string, line: number, offset: number): Promise<void> {
        return this._makeTssRequest<void>("quickinfo", {
            file,
            line,
            offset,
        })
    }

    public saveTo(file: string, tmpfile: string): Promise<void> {
        return this._makeTssRequest<void>("saveto", {
            file,
            tmpfile,
        })
    }

    public getSignatureHelp(file: string, line: number, offset: number): Promise<any> {
        return this._makeTssRequest<void>("signatureHelp", {
            file,
            line,
            offset,
        })
    }

    public getErrors(fullFilePath: string): Promise<void> {
        return this._makeTssRequest<void>("geterr", {
            files: [fullFilePath],
        })
    }

    public getErrorsAcrossProject(fullFilePath: string): Promise<void> {
        return this._makeTssRequest<void>("geterrForProject", {
            file: fullFilePath,
        })
    }

    public getNavigationTree(fullFilePath: string): Promise<INavigationTree> {
        return this._makeTssRequest<INavigationTree>("navtree", {
            file: fullFilePath,
        })
    }

    public getDocumentHighlights(file: string, line: number, offset: number): Promise<void> {
        return this._makeTssRequest<void>("documentHighlights", {
            file,
            line,
            offset,
        })
    }

   public _makeTssRequest<T>(commandName: string, args: any): Promise<T> {
        const seq = this._seqNumber++
        const payload = {
            seq,
            type: "request",
            command: commandName,
            arguments: args,
        }

        const ret = this._createDeferredPromise<T>()
        this._seqToPromises[seq] = ret

        // TODO: Handle updates in parallel?

        this._tssProcess.stdin.write(JSON.stringify(payload) + os.EOL)

        return ret.promise
    }

    private _parseResponse(returnedData: string): void {
        const response = JSON.parse(returnedData)

        const seq = response["request_seq"] // tslint:disable-line no-string-literal
        const success = response["success"] // tslint:disable-line no-string-literal

        if (typeof seq === "number") {
            if (success) {
                this._seqToPromises[seq].resolve(response.body)
            } else {
                this._seqToPromises[seq].reject(new Error(response.message))
            }
        } else {
            // If a sequence wasn't specified, it might be a call that returns multiple results
            // Like 'geterr' - returns both semanticDiag and syntaxDiag
            console.log("No sequence number returned.") // tslint:disable-line no-console

            if (response.type && response.type === "event") {
                if (response.event && response.event === "semanticDiag") {
                    this.emit("semanticDiag", response.body)
                }
            }
        }
    }

    private _createDeferredPromise<T>(): any {
        let resolve: Function
        let reject: Function
        const promise = new Promise((res, rej) => {
            resolve = res
            reject = rej
        })
        return {
            resolve,
            reject,
            promise,
        }
    }

}
